
/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * IT'S CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Module definition information - the part between the -START and -END
 * lines below is used by Configure. This could be stored in a separate
 * instead.
 *
 * MODULE-DEFINITION-START
 * Name: pointer_dbm_module
 * ConfigStart
    . ./helpers/find-dbm-lib
 * ConfigEnd
 * MODULE-DEFINITION-END
 */

/*
 * $Id: mod_dbm_pointer.c,v 1.6 2002/06/18 13:57:56 thomas Exp $
 * 
 * mod_dbm_pointer redirects whole domains based on records that are present 
 * in a dbm file.
 *
 * Creator: Thomas Eibner, <thomas@stderr.net>
 * Based on Vivek Khera's mod_auth_mysql.c (ftp://ftp.kciLink.com/pub/)
 * and on mod_auth_dbm.c with the standard 1.3 distribution
 *
 * Configuration: All configuration is done in the main configuration
 * file.
 * 
 * Enabling mod_pointer (defaults to Off):
 * PointerEnable On
 * Turns on pointer redirects for whole server, so be carefull on
 * checking what is written to the db.
 * 
 * Pointer redirects dbm file (defaults to NULL):
 * PointerDBMFile /usr/local/apache/conf/pointers
 *
 */

#include "mod_pointer.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#if (defined(NETWARE)|| defined(WIN32))
#define HAVE_SDBM
#endif
#ifdef HAVE_SDBM
#include <sdbm.h>
#define dbm_open sdbm_open
#define dbm_fetch sdbm_fetch
#define dbm_close sdbm_close
#else
#include <ndbm.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
    char *dbm_file;
    int pointerenable;
} pointer_server_cfg;

module MODULE_VAR_EXPORT pointer_dbm_module;

static void *pointer_server_config(pool *p, server_rec *s) {
    pointer_server_cfg *cfg = ap_pcalloc(p, sizeof(pointer_server_cfg));
    if (!cfg)
        return NULL;

    cfg->dbm_file = NULL;
    cfg->pointerenable = 0;
    return (void *)cfg;
}

static const char *pointer_set_pointerenable(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_dbm_module);
    cfg->pointerenable = flag;
    return NULL;
}

static const char *pointer_set_dbm_file(cmd_parms *cmd, void *dummy, char *dbm_file) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_dbm_module);
    cfg->dbm_file = dbm_file;
    return NULL;
}

static command_rec pointer_cmds[] = {
    { "PointerDBMFile", pointer_set_dbm_file, NULL,
      RSRC_CONF, TAKE1, "pointer dbm file" },
    { "PointerEnable", pointer_set_pointerenable, NULL,
      RSRC_CONF, FLAG, "turn on mod_pointer" },
    { NULL }
};

static int pointer_redirect(request_rec *r) {
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_dbm_module);
    char *redirect_url;
    DBM *f;
    datum d, q;

    if (!cfg->pointerenable) 
        return DECLINED;

    if (!r->hostname)
        return DECLINED;

    if (cfg->dbm_file == NULL)
        return DECLINED;

    if (!(f = dbm_open(cfg->dbm_file, O_RDONLY, 0644))) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Error while opening dbm file");
        return DECLINED;
    }

    q.dptr = ap_pstrdup(r->pool, r->hostname);
    q.dsize = strlen(q.dptr);

    d = dbm_fetch(f, q);

    if (d.dptr != NULL) {
        redirect_url = ap_pstrdup(r->pool, d.dptr);
    } else {
        redirect_url = "";
    }
    dbm_close(f);

    if (strcmp(redirect_url, "") != 0) {
        ap_table_setn(r->headers_out, "Location", redirect_url);
        ap_table_setn(r->err_headers_out, "X-Redirected-By", "mod_pointer - http://stderr.net/mod_pointer/");
        return REDIRECT;
    }
    ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: wtf?");
    return DECLINED;
}

module MODULE_VAR_EXPORT pointer_dbm_module = {
   STANDARD_MODULE_STUFF,
   pointer_server_init, /* initializer */
   NULL,			/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   pointer_server_config, /* server config */ 
   NULL,			/* merge server config */
   pointer_cmds,	/* command table */
   NULL,            /* handlers */
   NULL,			/* filename translation */
   NULL,            /* check_user_id */
   NULL,		    /* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL,			/* header parser */
   NULL,			/* child_init */
   NULL,		    /* child_exit */
   pointer_redirect /* post read-request */
};
