
/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * IT'S CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Module definition information - the part between the -START and -END
 * lines below is used by Configure. This could be stored in a separate
 * instead.
 *
 * MODULE-DEFINITION-START
 * Name: pointer_my_module
 * ConfigStart
     MYSQL_LIB="-L/usr/local/lib/mysql -lmysqlclient -lm"
     if [ "X$MYSQL_LIB" != "X" ]; then
         LIBS="$LIBS $MYSQL_LIB"
         echo " + using $MYSQL_LIB for MySQL support"
     fi
 * ConfigEnd
 * MODULE-DEFINITION-END
 */

/*
 * $Id: mod_my_pointer.c,v 1.12 2002/06/18 15:33:13 thomas Exp $
 * 
 * mod_my_pointer redirects whole domains based on records that are present 
 * in a MySQL database.
 *
 * Creator: Thomas Eibner, <thomas@stderr.net>
 * Based on Vivek Khera's mod_auth_mysql.c (ftp://ftp.kciLink.com/pub/)
 *
 * Configuration: All configuration is done in the main configuration
 * file.
 * 
 * Enabling mod_pointer (defaults to Off):
 * PointerEnable On
 * Turns on pointer redirects for whole server, so be carefull on
 * checking what is written to the db.
 * 
 * Setting MySQL host (port) (defaults to localhost and 3306):
 * PointerMySQLHost localhost 3306
 *
 * Setting MySQL user (defaults to pointer):
 * PointerMySQLUser pointer 
 *
 * Setting MySQL password (defaults to retniop):
 * PointerMySQLPassword retniop
 *
 * Setting MySQL DB (defaults to pointer):
 * PointerMySQLDB pointer
 *
 * Setting MySQL table (defaults to pointers):
 * PointerMySQLTable pointers
 *
 * Setting MySQL host field (defaults to host):
 * PointerMySQLHostField host
 *
 * Setting MySQL url field (defaults to url):
 * PointerMySQLURLField url
 *
 * Setting MySQL pass args field (defaults to pass_args):
 * PointerMySQLPassArgsField pass_args
 *
 * Enabling Accounting (defaults to Off):
 * PointerAccountingEnable On
 *
 * Setting MySQL accounting field (defaults to account):
 * PointerMySQLAccountField account
 *
 * Enabling persistent MySQL connections (defaults to Off):
 * PointerMySQLKeepAlive On
 *
 * Setting the frameset file that mod_pointer uses for framed redirects
 * PointerFramesetFile /path/to/file
 */

#include "mod_pointer.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include <mysql/mysql.h>
#include "buf_io.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define MAX_QUERY_LENGTH 300
#define MAX_HOST_LENGTH 255

typedef struct {
    char *mysqlhost;
    char *mysqluser;
    char *mysqlpasswd;
    char *mysqldb;
    char *mysqlptrtable;
    char *mysqlhostfield;
    char *mysqlurlfield;
    char *mysqlaccountfield;
    char *mysqlframefield;
    char *mysqltitlefield;
    char *mysqlpassargsfield;
    char *framesetfile;
    int mysqlport;
    int pointerenable;
    int accountingenable;
    int mysqlkeepalive;
} pointer_server_cfg;

static MYSQL *mysql_handle = NULL;

module MODULE_VAR_EXPORT pointer_my_module;

static void pointer_mysql_cleanup(void *notused) {
    if (mysql_handle)
        mysql_close(mysql_handle);
    mysql_handle = NULL;
}

static void pointer_mysql_cleanup_child(void *notused) {
    /* nothing */
}

static void child_exit(server_rec *s, pool *p) {
    pointer_mysql_cleanup(NULL);
}

static int open_db_handle(request_rec *r, pointer_server_cfg *cfg) {
    char *db_host;

    /* pinging would be better than just looking for wheter the handle is not NULL */
    if (mysql_handle && mysql_ping(mysql_handle) == 0)
        return TRUE;

    if (!cfg->mysqlhost || strcmp(cfg->mysqlhost, "localhost") == 0) {
        db_host = NULL;
    } else {
        db_host = cfg->mysqlhost;
    }

    mysql_handle = mysql_init(mysql_handle);

    if (mysql_real_connect(mysql_handle, db_host, cfg->mysqluser, cfg->mysqlpasswd, cfg->mysqldb, cfg->mysqlport, NULL, 0) == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: MySQL error: %s", mysql_error(mysql_handle));
    }

    if (!cfg->mysqlkeepalive) {
            /* close when request done */
        ap_register_cleanup(r->pool, (void *)NULL,
                            pointer_mysql_cleanup,
                            pointer_mysql_cleanup_child);
    }        

    return TRUE;
}

static void *pointer_server_config(pool *p, server_rec *s) {
    pointer_server_cfg *cfg = ap_pcalloc(p, sizeof(pointer_server_cfg));
    if (!cfg)
        return NULL;

    cfg->mysqlhost = "localhost";
    cfg->mysqluser = "pointer";
    cfg->mysqlpasswd = "retniop";
    cfg->mysqldb = "pointer";
    cfg->mysqlptrtable = "pointers";
    cfg->mysqlhostfield = "host";
    cfg->mysqlurlfield = "url";
    cfg->mysqlaccountfield = "account";
    cfg->mysqlframefield = "frame";
    cfg->mysqltitlefield = "title";
    cfg->mysqlpassargsfield = "pass_args";
    cfg->framesetfile = "/home/thomas/cvs/mod_pointer/frameset.html";
    cfg->mysqlport = 3306;
    cfg->mysqlkeepalive = 0;
    cfg->pointerenable = 0;
    cfg->accountingenable = 0;
    return (void *)cfg;
}

static const char *pointer_set_host(cmd_parms *cmd, void *dummy, char *host, char *port) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    if (port != NULL) {
        cfg->mysqlport = atoi(port);
    }
    cfg->mysqlhost = host;
    return NULL;
}

static const char *pointer_set_user(cmd_parms *cmd, void *dummy, char *user) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqluser = user;
    return NULL;
}

static const char *pointer_set_pass(cmd_parms *cmd, void *dummy, char *pass) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlpasswd = pass;
    return NULL;
}

static const char *pointer_set_db(cmd_parms *cmd, void *dummy, char *db) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqldb = db;
    return NULL;
}

static const char *pointer_set_table(cmd_parms *cmd, void *dummy, char *table) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlptrtable = table;
    return NULL;
}

static const char *pointer_set_host_field(cmd_parms *cmd, void *dummy, char *host_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlhostfield = host_field;
    return NULL;
}

static const char *pointer_set_url_field(cmd_parms *cmd, void *dummy, char *url_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlurlfield = url_field;
    return NULL;
}

static const char *pointer_set_account_field(cmd_parms *cmd, void *dummy, char *account_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlaccountfield = account_field;
    return NULL;
}

static const char *pointer_set_passargs_field(cmd_parms *cmd, void *dummy, char *passargs_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlpassargsfield = passargs_field;
    return NULL;
}

static const char *pointer_set_framesetfile(cmd_parms *cmd, void *dummy, char *framesetfile) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->framesetfile = framesetfile;
    return NULL;
}

static const char *pointer_set_keepalive(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->mysqlkeepalive = flag;
    return NULL;
}

static const char *pointer_set_pointerenable(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->pointerenable = flag;
    return NULL;
}

static const char *pointer_set_accountenable(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_my_module);
    cfg->accountingenable = flag;
    return NULL;
}

/* pointer_set_(title|frame)_field */

static command_rec pointer_cmds[] = {
    { "PointerMySQLHost", pointer_set_host, NULL,
      RSRC_CONF, TAKE12, "mysql server hostname and port" },
    { "PointerMySQLUser", pointer_set_user, NULL,
      RSRC_CONF, TAKE1, "mysql username" },
    { "PointerMySQLPassword", pointer_set_pass, NULL,
      RSRC_CONF, TAKE1, "mysql password" },
    { "PointerMySQLDB", pointer_set_db, NULL,
      RSRC_CONF, TAKE1, "mysql database" },
    { "PointerMySQLTable", pointer_set_table, NULL,
      RSRC_CONF, TAKE1, "mysql pointer table" },
    { "PointerMySQLHostField", pointer_set_host_field, NULL,
      RSRC_CONF, TAKE1, "mysql host field" },
    { "PointerMySQLURLField", pointer_set_url_field, NULL,
      RSRC_CONF, TAKE1, "mysql url field" },
    { "PointerMySQLAccountField", pointer_set_account_field, NULL,
      RSRC_CONF, TAKE1, "mysql account field" },
    { "PointerMySQLPassArgsField", pointer_set_passargs_field, NULL,
      RSRC_CONF, TAKE1, "mysql passargs field" },
    { "PointerFramesetFile", pointer_set_framesetfile, NULL,
      RSRC_CONF, TAKE1, "pointer frameset file" },
    { "PointerMySQLKeepAlive", pointer_set_keepalive, NULL,
      RSRC_CONF, FLAG, "mysql connection kept open across requests if On" },
    { "PointerEnable", pointer_set_pointerenable, NULL,
      RSRC_CONF, FLAG, "turn on mod_pointer" },
    { "PointerAccountingEnable", pointer_set_accountenable, NULL,
      RSRC_CONF, FLAG, "turn on accounting for mod_pointer" },
    { NULL }
};

static int pointer_redirect(request_rec *r) {
    MYSQL_RES *result;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_my_module);
    char query[MAX_QUERY_LENGTH];
    char *redirect_url;
    char *pointer_host;
    int handling_request = 0;

    if (!cfg->pointerenable)
        return DECLINED;

    if (!r->hostname)
        return DECLINED;

    if (!open_db_handle(r, cfg)) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Error while opening db handle");
		return DECLINED;
    }

    pointer_host = (char*)ap_palloc(r->pool, MAX_HOST_LENGTH);

    if (pointer_host == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Could not allocate enough memory.");
        return DECLINED;
    }

    if (mysql_real_escape_string(mysql_handle, pointer_host, r->hostname, strlen(r->hostname)) >= MAX_HOST_LENGTH) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
                     "[mod_pointer]: Hostname %s too long: %s", r->hostname, r->uri);
        return DECLINED;
    }

	/* allocate memory for query instead of using an array? */

    ap_snprintf(query, sizeof(query)-1, "select %s,%s,%s,%s from %s WHERE %s = '%s'",
                cfg->mysqlurlfield, cfg->mysqlframefield, cfg->mysqltitlefield,
                cfg->mysqlpassargsfield, cfg->mysqlptrtable, cfg->mysqlhostfield, pointer_host);

    if (mysql_real_query(mysql_handle, query, MAX_QUERY_LENGTH) != 0) {
        ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
                      "[mod_pointer]: MySQL error %s: %s", mysql_error(mysql_handle), r->uri);
        return SERVER_ERROR;
    }

    /* first check for host in database and wheter it's a framed request */
    /* if it is a framed request fetch the title and the url of the remote site  */
    /* else just redirect request to url */
    /* select url,frame,title from pointers where host = 'r->hostname'; */
    
    result = mysql_store_result(mysql_handle);
    if (result && (mysql_num_rows(result) > 0)) {
        /* we just need the first matching row */
        int i = mysql_num_rows(result);
        if (i == 1) {
            MYSQL_ROW data = mysql_fetch_row(result);
            if (data[0]) {
                if (strcmp(data[3], "0") == 0) {
                    /* do not pass uri on.. */
                    redirect_url = ap_pstrdup(r->pool, data[0]);
                } else if (strcmp(data[3], "1") == 0) {
                    /* append uri to url and query string after that */
                    char *uri = r->uri;
                    /* remove the first / */
                    uri++;
                    /* query string? */
                    if (!r->args) {
                        redirect_url = ap_pstrcat(r->pool, data[0], uri, NULL);
                    } else {
                        redirect_url = ap_pstrcat(r->pool, data[0], uri, "?", r->args, NULL);
                    }
                } else if (strcmp(data[3], "2") == 0) {
                    /* append uri to url as QUERY_STRING */
                    redirect_url = ap_pstrcat(r->pool, data[0], "?", ap_escape_path_segment(r->pool, r->uri), NULL);
                }               
                if (strcmp(data[1], "1") == 0) {
                    ap_table_setn(r->notes, "pointer-handle", "1");
                    ap_table_set(r->notes, "pointer-url", redirect_url);
                    if (data[2] != NULL) {
                        ap_table_set(r->notes, "pointer-title", ap_escape_html(r->pool, data[2]));
                    }
                    redirect_url = "";
                }
            }
            handling_request = 1;
        } else {
            ap_log_error(APLOG_MARK, APLOG_ERR, r->server, 
                         "[mod_pointer]: More than one value for an unique key: %s", pointer_host);
            redirect_url = "";
        }
    } else {
        redirect_url = "";
    }

    if (result)
        mysql_free_result(result);

    /* redirect_url != 0|NULL instead of using strcmp */

    if (handling_request) {
        if (cfg->accountingenable) {
            ap_snprintf(query, sizeof(query)-1, "UPDATE %s SET %s = %s + 1 WHERE %s = '%s'",
                        cfg->mysqlptrtable, cfg->mysqlaccountfield, cfg->mysqlaccountfield,
                        cfg->mysqlhostfield, pointer_host);
            if (mysql_real_query(mysql_handle, query, MAX_QUERY_LENGTH) != 0) {
                ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
                             "[mod_pointer]: MySQL error %s: %s", mysql_error(mysql_handle), r->uri);
                return SERVER_ERROR;
            }

            result = mysql_store_result(mysql_handle);
            if (result && (mysql_num_rows(result) > 0)) {
                int i = mysql_num_rows(result);
                if (i != 1) {
                    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
                                 "[mod_pointer]: More than one value in update of accounting information: %s", r->uri);
                }
            }
            if (result)
                mysql_free_result(result);
        }
    }
    
    if (strcmp(redirect_url, "") != 0) {
        /* redirect to url and issue a Moved response */
        /* permanently moved perhaps? */
        ap_table_setn(r->headers_out, "Location", redirect_url);
        ap_table_setn(r->err_headers_out, "X-Redirected-By", "mod_pointer - http://stderr.net/mod_pointer/");
        return REDIRECT;
    }
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return OK;

    return DECLINED;
}

static int pointer_trans(request_rec *r) {
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_my_module);
    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;
    r->filename = ap_pstrdup(r->pool, cfg->framesetfile);
    return OK;
}

static int pointer_mime(request_rec *r) {
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_my_module);
    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;
    r->content_type = "text/pointer-parsed-frameset";
    return OK;
}

static int pointer_parser(request_rec *r) {
    FILE *f;
    struct buf_io *io;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_my_module);

    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;

    io = buf_init(r);
    if (io == 0) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: could not open file %s or allocate enough memory", r->filename);
        return HTTP_FORBIDDEN;
    }
     
    r->content_type = "text/html";

    ap_send_http_header(r);

    if (r->header_only) {
        buf_destroy(io);
        return OK;
    }

    parse_loop(io);
    buf_destroy(io);

    return OK;
}

static const handler_rec pointer_handlers[] = {
    {"text/pointer-parsed-frameset", pointer_parser},
    {NULL}
};

module MODULE_VAR_EXPORT pointer_my_module = {
   STANDARD_MODULE_STUFF,
   pointer_server_init, /* initializer */
   NULL,			/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   pointer_server_config, /* server config */ 
   NULL,			/* merge server config */
   pointer_cmds,	/* command table */
   pointer_handlers,/* handlers */
   pointer_trans,   /* filename translation */
   NULL,            /* check_user_id */
   NULL,		    /* check auth */
   NULL,			/* check access */
   pointer_mime,    /* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL,			/* header parser */
   NULL,			/* child_init */
   child_exit,		/* child_exit */
   pointer_redirect	/* post read-request */
};
