
/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * IT'S CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Module definition information - the part between the -START and -END
 * lines below is used by Configure. This could be stored in a separate
 * instead.
 *
 * MODULE-DEFINITION-START
 * Name: pointer_pg_module
 * ConfigStart
     PGSQL_LIB="-I/usr/include/postgresql -lpq -lm"
	 LIBS="$LIBS $PGSQL_LIB"
	 echo " + using $PGSQL_LIB for Postgresql support"
 * ConfigEnd
 * MODULE-DEFINITION-END
 */

/*
 * $Id: mod_pg_pointer.c,v 1.9 2002/06/18 13:57:56 thomas Exp $
 *
 * mod_pg_pointer redirects whole domains based on records that are present
 * in a Postgresql database.
 *
 * Configuration: All configuration is done in the main configuration
 * file.
 *
 * Enabling mod_pointer (defaults to Off):
 * PointerEnable On
 * Turns on pointer redirects for whole server, so be carefull on
 * checking what is written to the db.
 *
 * Setting Postgresql host (port) (defaults to localhost and 3306):
 * PointerPGSQLHost localhost 3306
 *
 * Setting Postgresql user (defaults to pointer):
 * PointerPGSQLUser pointer
 *
 * Setting Postgresql password (defaults to retniop):
 * PointerPGSQLPassword retniop
 *
 * Setting Postgresql DB (defaults to pointer):
 * PointerPGSQLDB pointer
 *
 * Setting Postgresql table (defaults to pointers):
 * PointerPGSQLTable pointers
 *
 * Setting Postgresql host field (defaults to host):
 * PointerPGSQLHostField host
 *
 * Setting Postgresql url field (defaults to url):
 * PointerPGSQLURLField url
 *
 * Setting Postgresql pass args field (defaults to pass_args):
 * PointerPGSQLPassArgsField pass_args
 *
 * Enabling Accounting (defaults to Off):
 * PointerAccountingEnable On
 *
 * Setting Postgresql accounting field (defaults to account):
 * PointerPGSQLAccountField account
 *
 * Enabling persistent Postgresql connections (defaults to Off):
 * PointerPGSQLKeepAlive On
 *
 * Setting the frameset file that mod_pointer uses for framed redirects
 * PointerFramesetFile /path/to/file
 */

#include "mod_pointer.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include <libpq-fe.h>
#include "buf_io.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define MAX_QUERY_LENGTH 300
#define MAX_HOST_LENGTH 255

typedef struct {
    char *pgsqlhost;
    char *pgsqluser;
    char *pgsqlpasswd;
    char *pgsqldb;
    char *pgsqlptrtable;
    char *pgsqlhostfield;
    char *pgsqlurlfield;
    char *pgsqlaccountfield;
	/* yes, postgresql wants the port as a char and not an int, weird postgresql */
    char *pgsqlport;
    char *pgsqlframefield;
    char *pgsqltitlefield;
    char *pgsqlpassargsfield;
    char *framesetfile;
    int pointerenable;
    int accountingenable;
    int pgsqlkeepalive;
} pointer_server_cfg;

static PGconn *pgsql_handle = NULL;

module MODULE_VAR_EXPORT pointer_pg_module;

static int escape_string(char *to, const char *from, int length) {
    const char *to_start=to;
    const char *end;
    for (end=from+length; from != end; from++) {
        switch (*from) {
        case 0:
            *to++= '\\';
            *to++= '0';
            break;
        case '\\':
            *to++= '\\';
            *to++= '\\';
            break;
        case '\'':
            *to++= '\\';
            *to++= '\'';
            break;
        case '"':
            *to++= '\\';
            *to++= '"';
            break;
        default:
            *to++= *from;
        }
    }
    *to=0;
    return (int) (to-to_start);
}

static int pg_begin_transaction(request_rec *r, PGconn *pg_conn) {
	PGresult *res;

	res = PQexec(pg_conn, "BEGIN");
	if (!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: BEING transaction failed: %s", PQresultErrorMessage(res));
        PQclear(res);
		return 0;
	}
	PQclear(res);

	return 1;
}

static int pg_end_transaction(request_rec *r, PGconn *pg_conn) {
	PGresult *res;

    res = PQexec(pg_conn, "COMMIT");
	if (!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: END (commit) transaction failed: %s", PQresultErrorMessage(res));
        PQclear(res);
		return 0;
	}
    PQclear(res);
	
	return 1;
}

static int pg_execute_query(request_rec *r, PGconn *pg_conn, char *query) {
	PGresult *res;

	if (!pg_begin_transaction(r, pg_conn)) {
		return 0;
	}
	
	res = PQexec(pg_conn, query);
	if (!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Update failed: %s", PQresultErrorMessage(res));
		PQclear(res);
		return 0;
	}
	PQclear(res);
	
	if (!pg_end_transaction(r, pgsql_handle)) {
		return 0;
	}
	
	return 1;
}

static void pointer_pgsql_cleanup(void *notused) {
    if (pgsql_handle)
        PQfinish(pgsql_handle);
    pgsql_handle = NULL;
}

static void pointer_pgsql_cleanup_child(void *notused) {
    /* nothing */
}

static void child_exit(server_rec *s, pool *p) {
    pointer_pgsql_cleanup(NULL);
}

static int open_db_handle(request_rec *r, pointer_server_cfg *cfg) {
    char *db_host;

	/* ping handle? */
    if (pgsql_handle != NULL)
		return TRUE;

    if (!cfg->pgsqlhost || strcmp(cfg->pgsqlhost, "localhost") == 0) {
        db_host = NULL;
    } else {
        db_host = cfg->pgsqlhost;
    }

    pgsql_handle = PQsetdbLogin(db_host, cfg->pgsqlport, NULL, NULL, cfg->pgsqldb, cfg->pgsqluser, cfg->pgsqlpasswd);

    if (PQstatus(pgsql_handle) != CONNECTION_OK) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: pgsql error: %s", PQerrorMessage(pgsql_handle));
        return FALSE;
    }

    if (!cfg->pgsqlkeepalive) {
            /* close when request done */
        ap_register_cleanup(r->pool, (void *)NULL,
                            pointer_pgsql_cleanup,
                            pointer_pgsql_cleanup_child);
    }        

    return TRUE;
}

static void *pointer_server_config(pool *p, server_rec *s) {
    pointer_server_cfg *cfg = ap_pcalloc(p, sizeof(pointer_server_cfg));
    if (!cfg)
        return NULL;

    cfg->pgsqlhost = NULL;
    cfg->pgsqluser = NULL;
    cfg->pgsqlpasswd = NULL;
    cfg->pgsqldb = "pointer";
    cfg->pgsqlptrtable = "pointers";
    cfg->pgsqlhostfield = "host";
    cfg->pgsqlurlfield = "url";
    cfg->pgsqlaccountfield = "account";
    cfg->pgsqlframefield = "frame";
    cfg->pgsqltitlefield = "title";
    cfg->pgsqlpassargsfield = "pass_args";
    cfg->framesetfile = "/home/thomas/cvs/mod_pointer/frameset.html";
    cfg->pgsqlport = NULL;
    cfg->pgsqlkeepalive = 0;
    cfg->pointerenable = 0;
    cfg->accountingenable = 0;
    return (void *)cfg;
}

static const char *pointer_set_host(cmd_parms *cmd, void *dummy, char *host, char *port) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    if (port != NULL) {
		cfg->pgsqlport = port;
    }
    cfg->pgsqlhost = host;
    return NULL;
}

static const char *pointer_set_user(cmd_parms *cmd, void *dummy, char *user) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqluser = user;
    return NULL;
}

static const char *pointer_set_pass(cmd_parms *cmd, void *dummy, char *pass) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlpasswd = pass;
    return NULL;
}

static const char *pointer_set_db(cmd_parms *cmd, void *dummy, char *db) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqldb = db;
    return NULL;
}

static const char *pointer_set_table(cmd_parms *cmd, void *dummy, char *table) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlptrtable = table;
    return NULL;
}

static const char *pointer_set_host_field(cmd_parms *cmd, void *dummy, char *host_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlhostfield = host_field;
    return NULL;
}

static const char *pointer_set_url_field(cmd_parms *cmd, void *dummy, char *url_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlurlfield = url_field;
    return NULL;
}

static const char *pointer_set_account_field(cmd_parms *cmd, void *dummy, char *account_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlaccountfield = account_field;
    return NULL;
}

static const char *pointer_set_passargs_field(cmd_parms *cmd, void *dummy, char *passargs_field) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlpassargsfield = passargs_field;
    return NULL;
}

static const char *pointer_set_framesetfile(cmd_parms *cmd, void *dummy, char *framesetfile) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->framesetfile = framesetfile;
    return NULL;
}

static const char *pointer_set_keepalive(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pgsqlkeepalive = flag;
    return NULL;
}

static const char *pointer_set_pointerenable(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->pointerenable = flag;
    return NULL;
}

static const char *pointer_set_accountenable(cmd_parms *cmd, void *dummy, int flag) {
    server_rec *s = cmd->server;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(s->module_config, &pointer_pg_module);
    cfg->accountingenable = flag;
    return NULL;
}


static command_rec pointer_cmds[] = {
    { "PointerPGSQLHost", pointer_set_host, NULL,
      RSRC_CONF, TAKE12, "pgsql server hostname and port" },
    { "PointerPGSQLUser", pointer_set_user, NULL,
      RSRC_CONF, TAKE1, "pgsql username" },
    { "PointerPGSQLPassword", pointer_set_pass, NULL,
      RSRC_CONF, TAKE1, "pgsql password" },
    { "PointerPGSQLDB", pointer_set_db, NULL,
      RSRC_CONF, TAKE1, "pgsql database" },
    { "PointerPGSQLTable", pointer_set_table, NULL,
      RSRC_CONF, TAKE1, "pgsql pointer table" },
    { "PointerPGSQLHostField", pointer_set_host_field, NULL,
      RSRC_CONF, TAKE1, "pgsql host field" },
    { "PointerPGSQLURLField", pointer_set_url_field, NULL,
      RSRC_CONF, TAKE1, "pgsql url field" },
    { "PointerPGSQLAccountField", pointer_set_account_field, NULL,
      RSRC_CONF, TAKE1, "pgsql account field" },
    { "PointerPGSQLPassArgsField", pointer_set_passargs_field, NULL,
      RSRC_CONF, TAKE1, "pgsql passargs field" },
    { "PointerFramesetFile", pointer_set_framesetfile, NULL,
      RSRC_CONF, TAKE1, "pointer frameset file" },
    { "PointerPGSQLKeepAlive", pointer_set_keepalive, NULL,
      RSRC_CONF, FLAG, "pgsql connection kept open across requests if On" },
    { "PointerEnable", pointer_set_pointerenable, NULL,
      RSRC_CONF, FLAG, "turn on mod_pointer" },
    { "PointerAccountingEnable", pointer_set_accountenable, NULL,
      RSRC_CONF, FLAG, "turn on accounting for mod_pointer" },
    { NULL }
};

static int pointer_redirect(request_rec *r) {
    PGresult *res;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_pg_module);
    char *query;
    char *redirect_url;
    char *tmp_redirect_url;
    char *pointer_host;
    char *frame;
    char *title;
    char *passargs;

    if (!cfg->pointerenable) 
        return DECLINED;

    if (!r->hostname)
        return DECLINED;

    if (!open_db_handle(r, cfg)) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Error while opening db handle");		
		return DECLINED;
    }

    pointer_host = (char*)ap_palloc(r->pool, MAX_HOST_LENGTH*2);
	
	if (pointer_host == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Could not allocate enough memory.");
		return DECLINED;
	}	

	if (escape_string(pointer_host, r->hostname, strlen(r->hostname)) >= MAX_HOST_LENGTH) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
                     "[mod_pointer]: Hostname %s too long for mod_pointer: %s", r->hostname, r->uri);
        return DECLINED;
	}

	query = (char*)ap_palloc(r->pool, MAX_QUERY_LENGTH);

	if (query == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: Could not allocate enough memory.");
		return DECLINED;
	}	

    ap_snprintf(query, MAX_QUERY_LENGTH-1, "DECLARE ptrcursor CURSOR FOR select %s,%s,%s,%s from %s WHERE %s = '%s'",
                cfg->pgsqlurlfield, cfg->pgsqlframefield, cfg->pgsqltitlefield,
                cfg->pgsqlpassargsfield, cfg->pgsqlptrtable, cfg->pgsqlhostfield, pointer_host);

	/* start query processing */

	if (!pg_begin_transaction(r, pgsql_handle)) {
		return DECLINED;
	}

	res = PQexec(pgsql_handle, query);
	if (!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: DECLARE cursor failed: %s", PQresultErrorMessage(res));
        PQclear(res);
		return DECLINED;
	}

	PQclear(res);

    res = PQexec(pgsql_handle, "FETCH 1 in ptrcursor");
    if (!res || PQresultStatus(res) != PGRES_TUPLES_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: FETCH 1 failed/no rows returned: %s", PQresultErrorMessage(res));
        PQclear(res);
		return DECLINED;
    }

	if (PQntuples(res) == 1) {
		tmp_redirect_url = ap_pstrdup(r->pool, (char*)PQgetvalue(res, 0, 0));
        frame = ap_pstrdup(r->pool, (char*)PQgetvalue(res, 0, 1));
        title = ap_pstrdup(r->pool, (char*)PQgetvalue(res, 0, 2));
        passargs = ap_pstrdup(r->pool, (char*)PQgetvalue(res, 0, 3));
	} else {
		tmp_redirect_url = NULL;
	}

	PQclear(res);

    res = PQexec(pgsql_handle, "CLOSE ptrcursor");
	if (!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: DECLARE cursor failed: %s", PQresultErrorMessage(res));
        PQclear(res);
		return DECLINED;
	}

    PQclear(res);

	if (!pg_end_transaction(r, pgsql_handle)) {
		return DECLINED;
	}

    if (tmp_redirect_url != NULL) {
        if (cfg->accountingenable) {
            ap_snprintf(query, MAX_QUERY_LENGTH-1, "UPDATE %s SET %s = %s + 1 WHERE %s = '%s'",
                        cfg->pgsqlptrtable, cfg->pgsqlaccountfield, cfg->pgsqlaccountfield,
                        cfg->pgsqlhostfield, pointer_host);
			

			pg_execute_query(r, pgsql_handle, query);

		}

        if (strcmp(passargs, "0") == 0) {
            /* do not pass uri on.. */
            redirect_url = tmp_redirect_url;
        } else if (strcmp(passargs, "1") == 0) {
            /* append uri to url and query string after that */
            char *uri = r->uri;
            /* remove the first / */
            uri++;
            /* append query string if present */
            if (!r->args) {
                redirect_url = ap_pstrcat(r->pool, tmp_redirect_url, uri, NULL);
            } else {
                redirect_url = ap_pstrcat(r->pool, tmp_redirect_url, uri, "?", r->args, NULL);
            }
        } else if (strcmp(passargs, "2") == 0) {
            /* append uri to url as QUERY_STRING */
            redirect_url = ap_pstrcat(r->pool, tmp_redirect_url, "?", ap_escape_path_segment(r->pool, r->uri), NULL);
        }
        if (strcmp(frame, "1") == 0) {
            /* we are using a frame */
            ap_table_setn(r->notes, "pointer-handle", "1");
            ap_table_set(r->notes, "pointer-url", redirect_url);
            ap_table_set(r->notes, "pointer-title", ap_escape_html(r->pool, title));
            return OK;
        } else {
            ap_table_setn(r->headers_out, "Location", redirect_url);
            ap_table_setn(r->err_headers_out, "X-Redirected-By", "mod_pointer - http://stderr.net/mod_pointer/");
            return REDIRECT;
        }		
	}
    return DECLINED;
}

static int pointer_trans(request_rec *r) {
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_pg_module);
    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;
    r->filename = ap_pstrdup(r->pool, cfg->framesetfile);
    return OK;
}

static int pointer_mime(request_rec *r) {
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_pg_module);
    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;
    r->content_type = "text/pointer-parsed-frameset";
    return OK;
}

static int pointer_parser(request_rec *r) {
    FILE *f;
    struct buf_io *io;
    pointer_server_cfg *cfg = (pointer_server_cfg *)ap_get_module_config(r->server->module_config, &pointer_pg_module);

    if (!cfg->pointerenable)
        return DECLINED;
    if (ap_table_get(r->notes, "pointer-handle") == NULL)
        return DECLINED;

    io = buf_init(r);
    if (io == 0) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "[mod_pointer]: could not open file %s or allocate enough memory", r->filename);
        return HTTP_FORBIDDEN;
    }
     
    r->content_type = "text/html";

    ap_send_http_header(r);

    if (r->header_only) {
        buf_destroy(io);
        return OK;
    }

    parse_loop(io);
    buf_destroy(io);

    return OK;
}


static const handler_rec pointer_handlers[] = {
    {"text/pointer-parsed-frameset", pointer_parser},
    {NULL}
};

module MODULE_VAR_EXPORT pointer_pg_module = {
   STANDARD_MODULE_STUFF,
   pointer_server_init, /* initializer */
   NULL,			/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   pointer_server_config, /* server config */ 
   NULL,			/* merge server config */
   pointer_cmds,	/* command table */
   pointer_handlers,/* handlers */
   pointer_trans,   /* filename translation */
   NULL,            /* check_user_id */
   NULL,		    /* check auth */
   NULL,			/* check access */
   pointer_mime,    /* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL,			/* header parser */
   NULL,			/* child_init */
   child_exit,		/* child_exit */
   pointer_redirect	/* post read-request */
};
